#!/usr/bin/perl -W
#
#	This script is to discern the topologically equivalent atoms
#	in the molecule to use in conjunction with RESP program
#

sub usage {
die <<USAGE
   Usage:  $0  <mol.hin>  <Qtot>

Generate input RESP files for both stages 1 and 2.
On output mol.st1 and mol.st2
Alternatively generate input for bare ESP-fit charge calculation.
On output mol.st0

USAGE
}

usage unless $#ARGV == 1;

sub read_str;
sub read_str_from_HIN;
sub read_str_from_MOL2{ 0 };	# unsupported yet

sub discern_symm;
sub is_atom_same;

# wrappers to call
sub gen_stage0;
sub gen_stage1;
sub gen_stage2;

my %el2atw = (
"XX"  =>   0, "H"   =>   1, "HE"  =>   2,  
"LI"  =>   3, "BE"  =>   4, "B"   =>   5, "C"   =>   6, "N"   =>   7, "O"   =>   8, "F"   =>   9, "NE"  =>  10, 
"NA"  =>  11, "MG"  =>  12, "AL"  =>  13, "SI"  =>  14, "P"   =>  15, "S"   =>  16, "CL"  =>  17, "AR"  =>  18, 
"K"   =>  19, "CA"  =>  20, "SC"  =>  21, "TI"  =>  22, "V"   =>  23, "CR"  =>  24, "MN"  =>  25, "FE"  =>  26, "CO"  =>  27, "NI"  =>  28, "CU"  =>  29, "ZN"  =>  30, "GA"  =>  31, "GE"  =>  32, "AS"  =>  33, "SE"  =>  34, "BR"  =>  35, "KR"  =>  36, 
"RB"  =>  37, "SR"  =>  38, "Y"   =>  39, "ZR"  =>  40, "NB"  =>  41, "MO"  =>  42, "TC"  =>  43, "RU"  =>  44, "RH"  =>  45, "PD"  =>  46, "AG"  =>  47, "CD"  =>  48, "IN"  =>  49, "SN"  =>  50, "SB"  =>  51, "TE"  =>  52, "I"   =>  53, "XE"  =>  54, 
"CS"  =>  55, "BA"  =>  56, "LA"  =>  57, "CE"  =>  58, "PR"  =>  59, "ND"  =>  60, "PM"  =>  61, "SM"  =>  62, "EU"  =>  63, "GD"  =>  64, "TB"  =>  65, "DY"  =>  66, "HO"  =>  67, "ER"  =>  68, "TM"  =>  69, "YB"  =>  70, "LU"  =>  71, "HF"  =>  72, "TA"  =>  73, "W"   =>  74, "RE"  =>  75, "OS"  =>  76, "IR"  =>  77, "PT"  =>  78, "AU"  =>  79, "HG"  =>  80, "TL"  =>  81, "PB"  =>  82, "BI"  =>  83, "PO"  =>  84, "AT"  =>  85, "RN"  =>  86, 
"FR"  =>  87, "RA"  =>  88, "AC"  =>  89, "TH"  =>  90, "PA"  =>  91, "U"   =>  92, "NP"  =>  93, "PU"  =>  94, "AM"  =>  95, "CM"  =>  96, "BK"  =>  97, "CF"  =>  98, "ES"  =>  99, "FM"  => 100, "MD"  => 101, "NO"  => 102, "LR"  => 103, "RF"  => 104, "DB"  => 105, "SG"  => 106, "BH"  => 107, "HS"  => 108, "MT"  => 109, "DS"  => 110, "RG"  => 111, "UUB" => 112, "UUT" => 113, "UUQ" => 114, "UUP" => 115, "UUH" => 116
);

local $binm = 0;	#depends on version of resp program used: 0xd0xa/0xa - "\n"

my (%str, %sym);
my $qtot = $ARGV[1];

if(!read_str($ARGV[0], \%str)){
	die "error: couldn't read molecule structure\n";
}
	foreach my $k (sort {$a <=> $b} keys %str){
		print join(" ", @{$str{$k}}) . "\n";
	}

if(!discern_symm(\%str, \%sym)){
	die "error: couldn't discern topological symmetry\n";
}

# printing results
foreach my $k (sort {$a <=> $b} keys %sym){
	print "$k   ";
	if($sym{$k} == $k){
		print "0\n";
	} else {
		print "$sym{$k}\n";
	}
#	print join(" ", $sym{$k}) . "\n";
}

my ($bn, $st0, $st1, $st2, @res);
($bn = $ARGV[0]) =~ s/\.([^\.]+)$//;
$st0 = "$bn.st0";
$st1 = "$bn.st1";
$st2 = "$bn.st2";

# bare ESP-fit without symmetry consideration and weighing
print "\nStage 0\n";
if(gen_stage0(\%str, $st0, $qtot, \@res)){
	open ST0, ">$st0" or die "cannot open file $st0 for output: $!";
	binmode ST0 if $binm;
	print ST0 join("\n", @res);
	close ST0;
} else {
	die "couldn't generate stage0 input";
}


# RESP stages output
if(!discern_symm(\%str, \%sym, \&is_atom_same_st1)){
	die "error: couldn't discern stage1 symmetry\n";
}

# printing results
print "\nStage 1\n";
foreach my $k (sort {$a <=> $b} keys %sym){
	print "$k   ";
	if($sym{$k} == $k){
		print "0\n";
	} else {
		print "$sym{$k}\n";
	}
#	print join(" ", $sym{$k}) . "\n";
}

if(gen_stage1(\%str, \%sym, $st1, $qtot, \@res)){
	open ST1, ">$st1" or die "cannot open file $st1 for output: $!";
	binmode ST1 if $binm;
	print ST1 join("\n", @res);
	close ST1;
} else {
	die "couldn't generate stage1 input";
}


if(!discern_symm(\%str, \%sym, \&is_atom_same_st2)){
	die "error: couldn't discern stage1 symmetry\n";
}

# printing results
print "\nStage 2\n";
foreach my $k (sort {$a <=> $b} keys %sym){
	print "$k   ";
	if($sym{$k} == $k){
		if (@{$str{$k}}[4] > 0){
			# refit
			print " 0\n";
		} else {
			# freeze
			print "-1\n"
		}
	} else {
		print "$sym{$k}\n";
	}
#	print join(" ", $sym{$k}) . "\n";
}
if(gen_stage2(\%str, \%sym, $st1, $qtot, \@res)){
	open ST2, ">$st2" or die "cannot open file $st2 for output: $!";
	binmode ST2 if $binm;
	print ST2 join("\n", @res);
	close ST2;
} else {
	die "couldn't generate stage2 input";
}



#	Read structure from one of the supported file formats
#	Call syntax:	read_str([in] $file_name, [out] \%str)
sub read_str{
	return undef unless $#_ >= 1;
	my ($hfile) = shift @_;
	
	open F, "<$hfile" or die "cannot open file $hfile: $!";
	read_str_from_HIN ( F, @_ ) if $hfile =~ /\.hin$|\.hh$/;
	read_str_from_MOL2( F, @_ ) if $hfile =~ /\.ml2$|\.mol2$/;
	# ..., etc.
	close F;
	1;
}

#	Read HIN structure
#	Call syntax:	read_str_from_HIN([in] \FH, [out] \%str)
sub read_str_from_HIN{
	return undef unless $#_ >= 1;
	my ($fh, $str) = @_[0..1];
	my ($s, %bonds);
	while($s = <$fh>){
		last if $s =~ /^endmol/;
		next unless $s =~ /^atom/;
		my @ss = split(/\s+/, $s);
		warn "bad atom $ss[1]\n" if $#ss < 12;
#		push @{$$str{$ss[1]}}, (@ss[1,3,6,10], 0);	# atid, Elem, chg, nbonds, CHn
		push @{$$str{$ss[1]}}, ($ss[1], uc($ss[3]), @ss[6,10], 0);	# atid, Elem, chg, nbonds, CHn
		my @t = @ss[11 .. $#ss ];
		my @p;
		while(@p[0,1] = splice (@t, 0, 2)){
			push @{$bonds{$ss[1]}}, $p[0];
#			print "$p[0] ";
		}
	}
	my @el = grep { @{$$str{$_}}[1] =~ /^C$/ } keys %bonds;
	foreach my $j (@el){
		my ($ch, $lk) = 0;
		foreach my $k (@{$bonds{$j}}){
			if (@{$$str{$k}}[1] eq 'H') {
				++$ch;
				@{$$str{$k}}[4] = 1;	# label H's from CHn
				$lk = $k;
			};
		}
		if($ch == 1){
			@{$$str{$lk}}[4] = 0;
		} else {
			@{$$str{$j}}[4] = $ch;
		}
	}
	return scalar(%$str) ? 1 : undef;
}

#	Discern symmetry based on topological charge values (NB: not exact!)
#	Call syntax:	discern_symm(\%str, \%sym, \&is_atom_same)
sub discern_symm{
	return undef unless $#_ >=1;
	my ($str, $sym, $issame) = (@_[0..1], \&is_atom_same);
	if ($#_ >=2){
		$issame = $_[2];
	}
	# init
	my @k = sort {$a <=> $b} keys %$str;
# print join(" ", @k) . "\n";
	foreach my $i (@k){
#		push @{$$sym{$i}}, $i;
		$$sym{$i} = $i;
	}
	
	# main cycle
#	foreach my $i (@k){
#	}
	my $cur;
	while ($cur = pop @k){
		foreach my $i (@k){
			if(&$issame($str, $cur, $i)){
#				push @{$$sym{$i}}, @{$$sym{$cur}};
				$$sym{$cur} =  $$sym{$i};
#				delete $$sym{$cur};
#				print "same: $cur $i\n";
				last;
			}
		}
	};

	return scalar(%$sym) ? 1 : undef;
}

#	Compare two atoms: whether they are equal?
#	Call syntax:	is_atom_same(\%str, $cur, $i)
sub is_atom_same{
	return undef unless $#_ >= 2;
	my ($str, $cur, $i) = @_[0..2];
	
	return undef if (@{$$str{$cur}}[1] ne @{$$str{$i}}[1]);	# same element
	my $t = @{$$str{$cur}}[2] - @{$$str{$i}}[2];
	$t = $t <.0 ? -$t : $t;
	return $t <= 0.00001;
}

#	Compare two atoms: whether they are equal?
#	Call syntax:	is_atom_same(\%str, $cur, $i)
sub is_atom_same_st1{
	return undef unless $#_ >= 2;
	my ($str, $cur, $i) = @_[0..2];
	
	return undef if (@{$$str{$cur}}[1] ne @{$$str{$i}}[1]);	# same element
	return undef if (@{$$str{$cur}}[4] > 0 or @{$$str{$i}}[4] > 0);	# -CH2- or -CH3
#	return undef if (@{$$str{$cur}}[1] eq 'H' or @{$$str{$i}}[1] eq 'H');	# not H
	my $t = @{$$str{$cur}}[2] - @{$$str{$i}}[2];
	$t = $t <.0 ? -$t : $t;
	return $t <= 0.00001;
}

#	Compare two atoms: whether they are equal?
#	Call syntax:	is_atom_same(\%str, $cur, $i)
sub is_atom_same_st2{
	return undef unless $#_ >= 2;
	my ($str, $cur, $i) = @_[0..2];
	
	return undef if (@{$$str{$cur}}[1] ne @{$$str{$i}}[1]);	# same element
	return undef unless (@{$$str{$cur}}[4] > 0 or @{$$str{$i}}[4] > 0);	# not -CH2- or -CH3
	my $t = @{$$str{$cur}}[2] - @{$$str{$i}}[2];
	$t = $t <.0 ? -$t : $t;
	return $t <= 0.00001;
}

#	Generate ESP stage0 input
#	Call syntax:	gen_stage1([in] \%str, [in] $name, [in] $chg, [out] \@res)
sub gen_stage0{
	return undef unless $#_ > 1;
	my ($str, $name, $chg, $res) = @_[0 .. 4];
	my $nat = scalar(keys %$str);
	@$res = ();	# clear
#	scalar(@$res) = $n;
	push @$res, "$name ESP fit stage 0";
	push @$res, " &cntrl";
	push @$res, " " . join(", ", qw/iqopt=1 nmol=1 ihfree=0 qwt=0.000/);
	push @$res, " &end";
	push @$res, " 1.0";	# relative weight
	push @$res, "$name";
	push @$res, sprintf("%5d%5d", $chg, $nat);
	
	#begin atoms
	foreach my $at (sort {$a <=> $b} keys %$str){
		my $el = @{$$str{$at}}[1];
		die "element $el not known" if ! exists ($el2atw{$el});
		push @$res, sprintf ("%5d%5d", $el2atw{$el}, 0);
	}
	push @$res, ("", "", "","");
	1;	
};


#	Generate RESP stage1 input
#	Call syntax:	gen_stage1([in] \%str, [in] \%sym, [in] $name, [in] $chg, [out] \@res)
sub gen_stage1{
	return undef unless $#_ > 1;
	my ($str, $sym, $name, $chg, $res) = @_[0 .. 4];
	my $nat = scalar(keys %$str);
	@$res = ();	# clear
#	scalar(@$res) = $n;
	push @$res, "$name RESP stage 1";
	push @$res, " &cntrl";
	push @$res, " " . join(", ", qw/iqopt=1 nmol=1 ihfree=0 qwt=0.0005/);
	push @$res, " &end";
	push @$res, " 1.0";	# relative weight
	push @$res, "$name";
	push @$res, sprintf("%5d%5d", $chg, $nat);
#	push @$res, sprintf("%5.2f%5d", $chg, $nat);
	
	#begin atoms
	foreach my $at (sort {$a <=> $b} keys %$str){
		my $el = @{$$str{$at}}[1];
		die "element $el not known" if ! exists ($el2atw{$el});
		if($$sym{$at} == $at){
#			push @$res, "\t" . $el2atw{$el} . "\t0";
			push @$res, sprintf ("%5d%5d", $el2atw{$el}, 0);
		} else {
#			print "$sym{$k}\n";
#			push @$res, "\t" . $el2atw{$el} . "\t" . $$sym{$at};
			push @$res, sprintf ("%5d%5d", $el2atw{$el}, $$sym{$at});
		}
	}
	push @$res, ("", "", "","");
	1;	
}

#	Generate RESP stage2 input
#	Call syntax:	gen_stage1([in] \%str, [in] \%sym, [in] $name, [in] $chg, [out] \@res)
sub gen_stage2{
	return undef unless $#_ > 1;
	my ($str, $sym, $name, $chg, $res) = @_[0 .. 4];
	my $nat = scalar(keys %$str);
	@$res = ();	# clear
#	scalar(@$res) = $n;
	push @$res, "$name RESP stage 2";
	push @$res, " &cntrl";
	push @$res, " " . join(", ", qw/iqopt=2 nmol=1 ihfree=0 qwt=0.001/);
	push @$res, " &end";
	push @$res, " 1.0";	# relative weight
	push @$res, "$name";
	push @$res, sprintf("%5d%5d", $chg, $nat);
#	push @$res, sprintf("%5f%5d", $chg, $nat);
	
	#begin atoms
	foreach my $at (sort {$a <=> $b} keys %$str){
		my $el = @{$$str{$at}}[1];
		my $out;
		die "element $el not known" if ! exists ($el2atw{$el});
		
		if($$sym{$at} == $at){
#			push @$res, "\t" . $el2atw{$el} . "\t0";
#			push @$res, sprintf ("%5d%5d", $el2atw{$el}, 0);
			if(@{$$str{$at}}[4]>0){
				$out = 0;
			} else {
				$out = -1;
			}
		} else {
#			print "$sym{$k}\n";
#			push @$res, "\t" . $el2atw{$el} . "\t" . $$sym{$at};
#			push @$res, sprintf ("%5d%5d", $el2atw{$el}, $$sym{$at});
			$out = $$sym{$at};
		}
		push @$res, sprintf ("%5d%5d", $el2atw{$el}, $out);

	}
	push @$res, ("", "", "","");
	1;	
}

=pod
sub gen_stage2;
		if($$sym{$at} == $at){
			if (@{$str{$k}}[4] > 0){
				# refit
				print " 0\n";
			} else {
				# freeze
				print "-1\n"
			}
		} else {
			print "$sym{$k}\n";
		}

=cut
